/* Native-dependent code for FreeBSD.

   Copyright (C) 2002, 2003, 2004 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.  */

#include "defs.h"
#include "gdbcore.h"
#include "inferior.h"
#include "regcache.h"
#include "regset.h"
#include "symfile.h"

#include "gdb_assert.h"
#include "gdb_string.h"
#include <sys/types.h>
#include <sys/procfs.h>
#include <sys/sysctl.h>
#include <sys/ptrace.h>

#include "elf-bfd.h"
#include "fbsd-nat.h"

void clear_step_resume_breakpoint (void);
void clear_step_resume_breakpoint_thread (void);
void (*reactivate_threads) (char*) = NULL;
void (*disable_threads) (void) = NULL;

static void (*mourn_inferior_beneath) (void);
static void (*detach_beneath) (char *args, int from_tty);
static ptid_t (*wait_beneath) (ptid_t ptid, 
			       struct target_waitstatus *ourstatus);
int follow_event_pid = 0;

/* Return a the name of file that can be opened to get the symbols for
   the child process identified by PID.  */

char *
fbsd_pid_to_exec_file (int pid)
{
  size_t len = MAXPATHLEN;
  char *buf = xcalloc (len, sizeof (char));
  char *path;

#ifdef KERN_PROC_PATHNAME
  int mib[4];

  mib[0] = CTL_KERN;
  mib[1] = KERN_PROC;
  mib[2] = KERN_PROC_PATHNAME;
  mib[3] = pid;
  if (sysctl (mib, 4, buf, &len, NULL, 0) == 0)
    return buf;
#endif

  path = xstrprintf ("/proc/%d/file", pid);
  if (readlink (path, buf, MAXPATHLEN) == -1)
    {
      xfree (buf);
      buf = NULL;
    }

  xfree (path);
  return buf;
}

static int
fbsd_read_mapping (FILE *mapfile, unsigned long *start, unsigned long *end,
		   char *protection)
{
  /* FreeBSD 5.1-RELEASE uses a 256-byte buffer.  */
  char buf[256];
  int resident, privateresident;
  unsigned long obj;
  int ret = EOF;

  /* As of FreeBSD 5.0-RELEASE, the layout is described in
     /usr/src/sys/fs/procfs/procfs_map.c.  Somewhere in 5.1-CURRENT a
     new column was added to the procfs map.  Therefore we can't use
     fscanf since we need to support older releases too.  */
  if (fgets (buf, sizeof buf, mapfile) != NULL)
    ret = sscanf (buf, "%lx %lx %d %d %lx %s", start, end,
		  &resident, &privateresident, &obj, protection);

  return (ret != 0 && ret != EOF);
}

/* Iterate over all the memory regions in the current inferior,
   calling FUNC for each memory region.  OBFD is passed as the last
   argument to FUNC.  */

int
fbsd_find_memory_regions (int (*func) (CORE_ADDR, unsigned long,
				       int, int, int, void *),
			  void *obfd)
{
  pid_t pid = PIDGET (inferior_ptid);
  char *mapfilename;
  FILE *mapfile;
  unsigned long start, end, size;
  char protection[4];
  int read, write, exec;

  mapfilename = xstrprintf ("/proc/%ld/map", (long) pid);
  mapfile = fopen (mapfilename, "r");
  if (mapfile == NULL)
    error (_("Couldn't open %s."), mapfilename);

  if (info_verbose)
    fprintf_filtered (gdb_stdout, 
		      "Reading memory regions from %s\n", mapfilename);

  /* Now iterate until end-of-file.  */
  while (fbsd_read_mapping (mapfile, &start, &end, &protection[0]))
    {
      size = end - start;

      read = (strchr (protection, 'r') != 0);
      write = (strchr (protection, 'w') != 0);
      exec = (strchr (protection, 'x') != 0);

      if (info_verbose)
	{
	  fprintf_filtered (gdb_stdout, 
			    "Save segment, %ld bytes at 0x%s (%c%c%c)\n", 
			    size, paddr_nz (start),
			    read ? 'r' : '-',
			    write ? 'w' : '-',
			    exec ? 'x' : '-');
	}

      /* Invoke the callback function to create the corefile segment. */
      func (start, size, read, write, exec, obfd);
    }

  fclose (mapfile);
  return 0;
}

/* Create appropriate note sections for a corefile, returning them in
   allocated memory.  */

char *
fbsd_make_corefile_notes (bfd *obfd, int *note_size)
{
  struct gdbarch *gdbarch = current_gdbarch;
  const struct regcache *regcache = current_regcache;
  gregset_t gregs;
  fpregset_t fpregs;
  char *note_data = NULL;
  Elf_Internal_Ehdr *i_ehdrp;
  const struct regset *regset;
  size_t size;

  /* Put a "FreeBSD" label in the ELF header.  */
  i_ehdrp = elf_elfheader (obfd);
  i_ehdrp->e_ident[EI_OSABI] = ELFOSABI_FREEBSD;

  gdb_assert (gdbarch_regset_from_core_section_p (gdbarch));

  size = sizeof gregs;
  regset = gdbarch_regset_from_core_section (gdbarch, ".reg", size);
  gdb_assert (regset && regset->collect_regset);
  regset->collect_regset (regset, regcache, -1, &gregs, size);

  note_data = elfcore_write_prstatus (obfd, note_data, note_size,
				      PIDGET (inferior_ptid),
				      stop_signal, &gregs);

  size = sizeof fpregs;
  regset = gdbarch_regset_from_core_section (gdbarch, ".reg2", size);
  gdb_assert (regset && regset->collect_regset);
  regset->collect_regset (regset, regcache, -1, &fpregs, size);

  note_data = elfcore_write_prfpreg (obfd, note_data, note_size,
				     &fpregs, sizeof (fpregs));

  if (get_exec_file (0))
    {
      char *fname = strrchr (get_exec_file (0), '/') + 1;
      char *psargs = xstrdup (fname);

      if (get_inferior_args ())
	psargs = reconcat (psargs, psargs, " ", get_inferior_args (), NULL);

      note_data = elfcore_write_prpsinfo (obfd, note_data, note_size,
					  fname, psargs);
    }

  make_cleanup (xfree, note_data);
  return note_data;
}

static ptid_t
fbsd_wait (ptid_t ptid, struct target_waitstatus *ourstatus)
{
  struct ptrace_lwpinfo lwpinfo;
  struct target_waitstatus stat;
  ptid_t ret;
  static ptid_t forking_child = {0,0,0};

  ret = wait_beneath (ptid, ourstatus);
#ifdef PL_FLAG_FORKED
  if (PIDGET (ret) >= 0 && ourstatus->kind == TARGET_WAITKIND_STOPPED &&
      (ourstatus->value.sig == TARGET_SIGNAL_TRAP ||
       ourstatus->value.sig == TARGET_SIGNAL_STOP) &&
      (ptrace(PT_LWPINFO, PIDGET (ret), (caddr_t)&lwpinfo, 
	      sizeof lwpinfo) == 0))
    {
      if (lwpinfo.pl_flags & PL_FLAG_CHILD)
	{
	  /* Leave the child in a stopped state until we get a fork event in 
	     the parent. That's when we decide which process to follow. */
	  ourstatus->kind = TARGET_WAITKIND_IGNORE;
	  forking_child = ret;
	}
      else if (lwpinfo.pl_flags & PL_FLAG_FORKED)
	{
	  /* We'd better be in the middle of processing a fork() event. */
	  gdb_assert (!ptid_equal (forking_child, null_ptid));
	  ourstatus->kind = TARGET_WAITKIND_FORKED;
	  ourstatus->value.related_pid = lwpinfo.pl_child_pid;
	  forking_child = null_ptid;
	}
      else if (lwpinfo.pl_flags & PL_FLAG_EXEC &&
	  PIDGET (ret) == follow_event_pid)
	{
	  ourstatus->kind = TARGET_WAITKIND_EXECD;
	  ourstatus->value.execd_pathname =
	    xstrdup (fbsd_pid_to_exec_file (PIDGET (ret)));
	}
    }
#endif

  return ret;
}

static void
fbsd_enable_event_reporting (int pid)
{
#ifdef PT_FOLLOW_FORK
  follow_event_pid = pid;
  if (ptrace(PT_FOLLOW_FORK, pid, (caddr_t) 0, 1) < 0)
    warning (_("Cannot follow fork on this target."));
#endif 
}

static void
fbsd_post_attach (int pid)
{
  fbsd_enable_event_reporting (pid);
}

static void
fbsd_post_startup_inferior (ptid_t ptid)
{
  fbsd_enable_event_reporting (PIDGET (ptid));
}

static int
fbsd_follow_fork (struct target_ops *ops, int follow_child)
{
  ptid_t last_ptid, ret, child_ptid;
  struct target_waitstatus last_status, ourstatus;
  int parent_pid, child_pid;

  get_last_target_status (&last_ptid, &last_status);
  parent_pid = PIDGET (last_ptid);
  child_pid = last_status.value.related_pid;

  if (follow_child)
    {
      detach_breakpoints (child_pid);
      remove_breakpoints ();
      child_ptid = pid_to_ptid (child_pid);

      target_detach (NULL, 0);
      inferior_ptid = child_ptid;

      /* Reinstall ourselves, since we might have been removed in
	 target_detach (which does other necessary cleanup).  */
      push_target (ops);

      /* Need to restore some of the actions done by the threaded detach */
      if (reactivate_threads) 
	{
	  reactivate_threads (fbsd_pid_to_exec_file (child_pid));
	  reactivate_threads = NULL;
	}

      /* Reset breakpoints in the child as appropriate.  */
      clear_step_resume_breakpoint_thread ();
      follow_inferior_reset_breakpoints ();

      /* Enable fork/exec event reporting for the child. */
      fbsd_enable_event_reporting (child_pid);
    }
  else /* Follow parent */
    {
      /* Before detaching from the child, remove all breakpoints from
         it.  (This won't actually modify the breakpoint list, but will
         physically remove the breakpoints from the child.) */
      detach_breakpoints (child_pid);
      ptrace (PT_DETACH, child_pid, (caddr_t) 1, 0);
    }

  return 0;
}

/* EXECD_PATHNAME is assumed to be non-NULL. */

static int
fbsd_follow_exec (int pid, char *execd_pathname)
{
  struct target_waitstatus status;
  ptid_t ret = inferior_ptid;

  /* This is an exec event that we actually wish to pay attention to.
     Refresh our symbol table to the newly exec'd program, remove any
     momentary bp's, etc.

     If there are breakpoints, they aren't really inserted now,
     since the exec() transformed our inferior into a fresh set
     of instructions.

     We want to preserve symbolic breakpoints on the list, since
     we have hopes that they can be reset after the new a.out's
     symbol table is read.

     However, any "raw" breakpoints must be removed from the list
     (e.g., the solib bp's), since their address is probably invalid
     now.

     And, we DON'T want to call delete_breakpoints() here, since
     that may write the bp's "shadow contents" (the instruction
     value that was overwritten witha TRAP instruction).  Since
     we now have a new a.out, those shadow contents aren't valid. */
  update_breakpoints_after_exec ();

  /* If there was one, it's gone now.  We cannot truly step-to-next
     statement through an exec(). */
  clear_step_resume_breakpoint ();
  step_range_start = 0;
  step_range_end = 0;

  /* What is this a.out's name? */
  printf_unfiltered (_("Executing new program: %s\n"), execd_pathname);

  /* We've followed the inferior through an exec.  Therefore, the
     inferior has essentially been killed & reborn. */
  gdb_flush (gdb_stdout);

  /* Disable thread library */
  if (disable_threads)
    {
      disable_threads ();
      disable_threads = NULL;
    }

  generic_mourn_inferior ();
  inferior_ptid = ret;

  /* That a.out is now the one to use. */
  exec_file_attach (execd_pathname, 0);

  /* And also is where symbols can be found. */
  symbol_file_add_main (execd_pathname, 0);

  /* Reset the shared library package.  This ensures that we get
     a shlib event when the child reaches "_start", at which point
     the dld will have had a chance to initialize the child. */
#if defined(SOLIB_RESTART)
  SOLIB_RESTART ();
#endif
#ifdef SOLIB_CREATE_INFERIOR_HOOK
  SOLIB_CREATE_INFERIOR_HOOK (PIDGET (inferior_ptid));
#else
  solib_create_inferior_hook ();
#endif

  /* Reinsert all breakpoints.  (Those which were symbolic have
     been reset to the proper address in the new a.out, thanks
     to symbol_file_command...) */
  insert_breakpoints ();

  return 0;
}

static void fbsd_mourn_inferior (void)
{
  follow_event_pid = 0;
  mourn_inferior_beneath ();
}

static void fbsd_detach (char *args, int from_tty)
{
  follow_event_pid = 0;
  detach_beneath (args, from_tty);
}

void
fbsd_init (struct target_ops *t)
{
  wait_beneath = t->to_wait;
  detach_beneath = t->to_detach;
  mourn_inferior_beneath = t->to_mourn_inferior;
  t->to_wait = fbsd_wait;
  t->to_detach = fbsd_detach;
  t->to_mourn_inferior = fbsd_mourn_inferior;
  t->to_post_attach = fbsd_post_attach;
  t->to_post_startup_inferior = fbsd_post_startup_inferior;
  t->to_follow_fork = fbsd_follow_fork;
  t->to_follow_exec = fbsd_follow_exec;
}
