/* $FreeBSD: src/gnu/usr.bin/gdb/arch/arm/armfbsd-tdep.c,v 1.2 2007/11/17 21:30:03 cognet Exp $ */

/* Target-specific functions for ARM running under FreeBSD.
   Copyright 2002, 2003 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "defs.h"
#include "osabi.h"

#include "arch-utils.h"
#include "arm-tdep.h"
#include "solib-svr4.h"

#include "inferior.h"
#include "regcache.h"
#include "gdbcore.h"
#include "bfd.h"
#include "elf-bfd.h"

extern int arm_apcs_32;

/* Description of the longjmp buffer.  */
#define ARM_FBSD_JB_PC 24
#define ARM_FBSD_JB_ELEMENT_SIZE INT_REGISTER_SIZE

static void
arm_freebsd_init_abi_common (struct gdbarch_info info,
			    struct gdbarch *gdbarch)
{
  struct gdbarch_tdep *tdep = gdbarch_tdep (gdbarch);

  tdep->lowest_pc = 0x8000;

  tdep->jb_pc = ARM_FBSD_JB_PC;
  tdep->jb_elt_size = ARM_FBSD_JB_ELEMENT_SIZE;
}
  
static void
arm_freebsd_elf_init_abi (struct gdbarch_info info, 
			 struct gdbarch *gdbarch)
{
  struct gdbarch_tdep *tdep = gdbarch_tdep (gdbarch);

  arm_freebsd_init_abi_common (info, gdbarch);

  /* FreeBSD ELF uses SVR4-style shared libraries.  */
  set_solib_svr4_fetch_link_map_offsets
    (gdbarch, svr4_ilp32_fetch_link_map_offsets);

  tdep->fp_model = ARM_FLOAT_SOFT_FPA;
}

typedef struct fp_extended_precision {
  u_int32_t fp_exponent;
  u_int32_t fp_mantissa_hi;
  u_int32_t fp_mantissa_lo;
} fp_extended_precision_t;

typedef struct fp_extended_precision fp_reg_t;

struct fpreg {
  unsigned int fpr_fpsr;
  fp_reg_t fpr[8];
};

void
supply_fpregset (struct fpreg *fparegset)
{
  int regno;

  for (regno = ARM_F0_REGNUM; regno <= ARM_F7_REGNUM; regno++)
    regcache_raw_supply (current_regcache, regno, (char *) &fparegset->fpr[regno - ARM_F0_REGNUM]);

  regcache_raw_supply (current_regcache, ARM_FPS_REGNUM, (char *) &fparegset->fpr_fpsr);
}

struct reg {
  unsigned int r[13];
  unsigned int r_sp;
  unsigned int r_lr;
  unsigned int r_pc;
  unsigned int r_cpsr;
};

void
supply_gregset (struct reg *gregset)
{
  int regno;
  CORE_ADDR r_pc;

  /* Integer registers.  */
  for (regno = ARM_A1_REGNUM; regno < ARM_SP_REGNUM; regno++)
    regcache_raw_supply (current_regcache, regno, (char *) &gregset->r[regno]);

  regcache_raw_supply (current_regcache, ARM_SP_REGNUM, (char *) &gregset->r_sp);
  regcache_raw_supply (current_regcache, ARM_LR_REGNUM, (char *) &gregset->r_lr);
  regcache_raw_supply (current_regcache, ARM_PC_REGNUM, (char *) &gregset->r_pc);

  if (arm_apcs_32)
    regcache_raw_supply (current_regcache, ARM_PS_REGNUM, (char *) &gregset->r_cpsr);
  else
    regcache_raw_supply (current_regcache, ARM_PS_REGNUM, (char *) &gregset->r_pc);
}

struct md_core
{
  struct reg intreg;
  struct fpreg freg;
};
  
static void
fetch_core_registers (char *core_reg_sect, unsigned core_reg_size,
                      int which, CORE_ADDR ignore)
{
  struct md_core *core_reg = (struct md_core *) core_reg_sect;
  int regno;
  CORE_ADDR r_pc;
  
  supply_gregset (&core_reg->intreg);
  supply_fpregset (&core_reg->freg);
}

static struct core_fns arm_freebsd_core_fns =
{
  bfd_target_unknown_flavour,           /* core_flovour.  */
  default_check_format,                 /* check_format.  */
  default_core_sniffer,                 /* core_sniffer.  */
  fetch_core_registers,                 /* core_read_registers.  */
  NULL
};

int
fbsd_core_sniffer (struct core_fns *our_fns, bfd *abfd)
{
  int result;

  result = (bfd_get_flavour (abfd) == our_fns -> core_flavour &&
            elf_elfheader(abfd)->e_ident[EI_OSABI] == ELFOSABI_FREEBSD);
  return (result);
}

static void
fetch_elfcore_registers (char *core_reg_sect, unsigned core_reg_size,
                         int which, CORE_ADDR ignore)
{
  struct reg gregset;
  struct fpreg fparegset;

  switch (which)
    {
    case 0:     /* Integer registers.  */
      if (core_reg_size != sizeof (struct reg))
        warning ("wrong size of register set in core file");
      else
        {
          /* The memcpy may be unnecessary, but we can't really be sure
             of the alignment of the data in the core file.  */
          memcpy (&gregset, core_reg_sect, sizeof (gregset));
          supply_gregset (&gregset);
        }
      break;

    case 2:
      if (core_reg_size != sizeof (struct fpreg))
        warning ("wrong size of FPA register set in core file");
      else
        {
          /* The memcpy may be unnecessary, but we can't really be sure
             of the alignment of the data in the core file.  */
          memcpy (&fparegset, core_reg_sect, sizeof (fparegset));
          supply_fpregset (&fparegset);
        }
      break;

    default:
      /* Don't know what kind of register request this is; just ignore it.  */
      break;
    }
}

static struct core_fns arm_freebsd_elfcore_fns =
{
  bfd_target_elf_flavour,               /* core_flovour.  */
  default_check_format,                 /* check_format.  */
  fbsd_core_sniffer,                 /* core_sniffer.  */
  fetch_elfcore_registers,              /* core_read_registers.  */
  NULL
};

void
_initialize_armfbsd_tdep (void)
{
  gdbarch_register_osabi (bfd_arch_arm, 0, GDB_OSABI_FREEBSD_ELF,
                          arm_freebsd_elf_init_abi);

  deprecated_add_core_fns (&arm_freebsd_elfcore_fns);
  deprecated_add_core_fns (&arm_freebsd_core_fns);
}
