/*
 * Copyright (c) 2004 Marcel Moolenaar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: src/gnu/usr.bin/gdb/kgdb/trgt_i386.c,v 1.2.2.5 2008/05/09 18:59:50 jhb Exp $");

#include <sys/param.h>
#include <sys/proc.h>
#include <machine/segments.h>
#include <err.h>
#include <kvm.h>
#include <string.h>

#include <defs.h>
#include <target.h>
#include <gdbthread.h>
#include <inferior.h>
#include <regcache.h>
#include <frame-unwind.h>
#include <i386-tdep.h>

#include "kgdb.h"

static size_t pcb_ebx_off, pcb_esp_off, pcb_ebp_off, pcb_esi_off, 
  pcb_edi_off, pcb_eip_off, pcb_size;

void
kgdb_trgt_fetch_registers(int regno __unused)
{
	struct kthr *kt;
	char *pcb;

	pcb = alloca(pcb_size);
	if (pcb == NULL)
		return;
	kt = kgdb_thr_lookup_tid(ptid_get_pid(inferior_ptid));
	if (kt == NULL)
		return;
	if (kvm_read(kvm, kt->pcb, pcb, pcb_size) != pcb_size) {
		warnx("kvm_read: %s", kvm_geterr(kvm));
		memset(pcb, 0, pcb_size);
	}

	regcache_raw_supply(current_regcache, I386_EBX_REGNUM, pcb + pcb_ebx_off);
	regcache_raw_supply(current_regcache, I386_ESP_REGNUM, pcb + pcb_esp_off);
	regcache_raw_supply(current_regcache, I386_EBP_REGNUM, pcb + pcb_ebp_off);
	regcache_raw_supply(current_regcache, I386_ESI_REGNUM, pcb + pcb_esi_off);
	regcache_raw_supply(current_regcache, I386_EDI_REGNUM, pcb + pcb_edi_off);
	regcache_raw_supply(current_regcache, I386_EIP_REGNUM, pcb + pcb_eip_off);
}

void
kgdb_trgt_store_registers(int regno __unused)
{
	fprintf_unfiltered(gdb_stderr, "XXX: %s\n", __func__);
}

struct kgdb_tss_cache {
	CORE_ADDR	pc;
	CORE_ADDR	sp;
	CORE_ADDR	tss;
};

static int kgdb_trgt_tss_offset[15];
static int kgdb_trgt_frame_offset[15];

void
kgdb_trgt_init_type_info(void)
{
	pcb_size = kgdb_parse("sizeof(struct pcb)");
	pcb_ebx_off = kgdb_parse("&((struct pcb *)0)->pcb_ebx");
	pcb_esp_off = kgdb_parse("&((struct pcb *)0)->pcb_esp");
	pcb_ebp_off = kgdb_parse("&((struct pcb *)0)->pcb_ebp");
	pcb_esi_off = kgdb_parse("&((struct pcb *)0)->pcb_esi");
	pcb_edi_off = kgdb_parse("&((struct pcb *)0)->pcb_edi");
	pcb_eip_off = kgdb_parse("&((struct pcb *)0)->pcb_eip");

	kgdb_trgt_tss_offset[0] = kgdb_parse("&((struct i386tss*)0)->tss_eax");
	kgdb_trgt_tss_offset[1] = kgdb_parse("&((struct i386tss*)0)->tss_ecx");
	kgdb_trgt_tss_offset[2] = kgdb_parse("&((struct i386tss*)0)->tss_edx");
	kgdb_trgt_tss_offset[3] = kgdb_parse("&((struct i386tss*)0)->tss_ebx");
	kgdb_trgt_tss_offset[4] = kgdb_parse("&((struct i386tss*)0)->tss_esp");
	kgdb_trgt_tss_offset[5] = kgdb_parse("&((struct i386tss*)0)->tss_ebp");
	kgdb_trgt_tss_offset[6] = kgdb_parse("&((struct i386tss*)0)->tss_esi");
	kgdb_trgt_tss_offset[7] = kgdb_parse("&((struct i386tss*)0)->tss_edi");
	kgdb_trgt_tss_offset[8] = kgdb_parse("&((struct i386tss*)0)->tss_eip");
	kgdb_trgt_tss_offset[9] = kgdb_parse("&((struct i386tss*)0)->tss_eflags");
	kgdb_trgt_tss_offset[10] = kgdb_parse("&((struct i386tss*)0)->tss_cs");
	kgdb_trgt_tss_offset[11] = kgdb_parse("&((struct i386tss*)0)->tss_ss");
	kgdb_trgt_tss_offset[12] = kgdb_parse("&((struct i386tss*)0)->tss_ds");
	kgdb_trgt_tss_offset[13] = kgdb_parse("&((struct i386tss*)0)->tss_es");
	kgdb_trgt_tss_offset[14] = kgdb_parse("&((struct i386tss*)0)->tss_fs");
 
	kgdb_trgt_frame_offset[0] = kgdb_parse("&((struct trapframe*)0)->tf_eax");
	kgdb_trgt_frame_offset[1] = kgdb_parse("&((struct trapframe*)0)->tf_ecx");
	kgdb_trgt_frame_offset[2] = kgdb_parse("&((struct trapframe*)0)->tf_edx");
	kgdb_trgt_frame_offset[3] = kgdb_parse("&((struct trapframe*)0)->tf_ebx");
	kgdb_trgt_frame_offset[4] = kgdb_parse("&((struct trapframe*)0)->tf_esp");
	kgdb_trgt_frame_offset[5] = kgdb_parse("&((struct trapframe*)0)->tf_ebp");
	kgdb_trgt_frame_offset[6] = kgdb_parse("&((struct trapframe*)0)->tf_esi");
	kgdb_trgt_frame_offset[7] = kgdb_parse("&((struct trapframe*)0)->tf_edi");
	kgdb_trgt_frame_offset[8] = kgdb_parse("&((struct trapframe*)0)->tf_eip");
	kgdb_trgt_frame_offset[9] = kgdb_parse("&((struct trapframe*)0)->tf_eflags");
	kgdb_trgt_frame_offset[10] = kgdb_parse("&((struct trapframe*)0)->tf_cs");
	kgdb_trgt_frame_offset[11] = kgdb_parse("&((struct trapframe*)0)->tf_ss");
	kgdb_trgt_frame_offset[12] = kgdb_parse("&((struct trapframe*)0)->tf_ds");
	kgdb_trgt_frame_offset[13] = kgdb_parse("&((struct trapframe*)0)->tf_es");
	kgdb_trgt_frame_offset[14] = kgdb_parse("&((struct trapframe*)0)->tf_fs");
}

/*
 * If the current thread is executing on a CPU, fetch the common_tss
 * for that CPU.
 *
 * This is painful because 'struct pcpu' is variant sized, so we can't
 * use it.  Instead, we lookup the GDT selector for this CPU and
 * extract the base of the TSS from there.
 */
static CORE_ADDR
kgdb_trgt_fetch_tss(void)
{
	struct kthr *kt;
	struct segment_descriptor sd;
	CORE_ADDR addr, cpu0prvpage, tss;

	kt = kgdb_thr_lookup_tid(ptid_get_pid(inferior_ptid));
	if (kt == NULL || kt->cpu == NOCPU)
		return (0);

	addr = kgdb_parse("&_gdt");
	if (addr == 0)
		return (0);
	addr += (kt->cpu * NGDT + GPROC0_SEL) * sizeof(sd);
	if (kvm_read(kvm, addr, &sd, sizeof(sd)) != sizeof(sd)) {
		warnx("kvm_read: %s", kvm_geterr(kvm));
		return (0);
	}
	if (sd.sd_type != SDT_SYS386BSY) {
		warnx("descriptor is not a busy TSS");
		return (0);
	}
	tss = sd.sd_hibase << 24 | sd.sd_lobase;

	/*
	 * In SMP kernels, the TSS is stored as part of the per-CPU
	 * data.  On older kernels, the CPU0's private page
	 * is stored at an address that isn't mapped in minidumps.
	 * However, the data is mapped at the alternate cpu0prvpage
	 * address.  Thus, if the TSS is at the invalid address,
	 * change it to be relative to cpu0prvpage instead.
	 */ 
	if (trunc_page(tss) == 0xffc00000) {
		addr = kgdb_parse("&_cpu0prvpage");
		if (addr == 0) {
			warnx("kgdb_parse(&_cpu0prvpage) failed");
			return (0);
		}
		if (kvm_read(kvm, addr, &cpu0prvpage, sizeof(cpu0prvpage)) !=
		    sizeof(cpu0prvpage)) {
			warnx("kvm_read: %s", kvm_geterr(kvm));
			return (0);
		}
		tss = cpu0prvpage + (tss & PAGE_MASK);
	}
	return ((CORE_ADDR)tss);
}

static struct kgdb_tss_cache *
kgdb_trgt_tss_cache(struct frame_info *next_frame, void **this_cache)
{
	char buf[MAX_REGISTER_SIZE];
	struct kgdb_tss_cache *cache;

	cache = *this_cache;
	if (cache == NULL) {
		cache = FRAME_OBSTACK_ZALLOC(struct kgdb_tss_cache);
		*this_cache = cache;
		cache->pc = frame_func_unwind(next_frame);
		frame_unwind_register(next_frame, SP_REGNUM, buf);
		cache->sp = extract_unsigned_integer(buf,
		    register_size(current_gdbarch, SP_REGNUM));
		cache->tss = kgdb_trgt_fetch_tss();
	}
	return (cache);
}

static void
kgdb_trgt_dblfault_this_id(struct frame_info *next_frame, void **this_cache,
    struct frame_id *this_id)
{
	struct kgdb_tss_cache *cache;

	cache = kgdb_trgt_tss_cache(next_frame, this_cache);
	*this_id = frame_id_build(cache->sp, cache->pc);
}

static void
kgdb_trgt_dblfault_prev_register(struct frame_info *next_frame,
    void **this_cache, int regnum, int *optimizedp, enum lval_type *lvalp,
    CORE_ADDR *addrp, int *realnump, void *valuep)
{
	char dummy_valuep[MAX_REGISTER_SIZE];
	struct kgdb_tss_cache *cache;
	int ofs, regsz;

	regsz = register_size(current_gdbarch, regnum);

	if (valuep == NULL)
		valuep = dummy_valuep;
	memset(valuep, 0, regsz);
	*optimizedp = 0;
	*addrp = 0;
	*lvalp = not_lval;
	*realnump = -1;

	ofs = (regnum >= I386_EAX_REGNUM && regnum <= I386_FS_REGNUM)
	    ? kgdb_trgt_tss_offset[regnum] : -1;
	if (ofs == -1)
		return;

	cache = kgdb_trgt_tss_cache(next_frame, this_cache);
	if (cache->tss == 0)
		return;
	*addrp = cache->tss + ofs;
	*lvalp = lval_memory;
	target_read_memory(*addrp, valuep, regsz);
}

static const struct frame_unwind kgdb_trgt_dblfault_unwind = {
        NORMAL_FRAME,
        &kgdb_trgt_dblfault_this_id,
        &kgdb_trgt_dblfault_prev_register
};

struct kgdb_frame_cache {
	int		frame_type;
	CORE_ADDR	pc;
	CORE_ADDR	sp;
};
#define	FT_NORMAL		1
#define	FT_INTRFRAME		2
#define	FT_INTRTRAPFRAME	3
#define	FT_TIMERFRAME		4

static struct kgdb_frame_cache *
kgdb_trgt_frame_cache(struct frame_info *next_frame, void **this_cache)
{
	char buf[MAX_REGISTER_SIZE];
	struct kgdb_frame_cache *cache;
	char *pname;

	cache = *this_cache;
	if (cache == NULL) {
		cache = FRAME_OBSTACK_ZALLOC(struct kgdb_frame_cache);
		*this_cache = cache;
		cache->pc = frame_func_unwind(next_frame);
		find_pc_partial_function(cache->pc, &pname, NULL, NULL);
		if (pname[0] != 'X')
			cache->frame_type = FT_NORMAL;
		else if (strcmp(pname, "Xtimerint") == 0)
			cache->frame_type = FT_TIMERFRAME;
		else if (strcmp(pname, "Xcpustop") == 0 ||
		    strcmp(pname, "Xrendezvous") == 0 ||
		    strcmp(pname, "Xipi_intr_bitmap_handler") == 0 ||
		    strcmp(pname, "Xlazypmap") == 0)
			cache->frame_type = FT_INTRTRAPFRAME;
		else
			cache->frame_type = FT_INTRFRAME;
		frame_unwind_register(next_frame, SP_REGNUM, buf);
		cache->sp = extract_unsigned_integer(buf,
		    register_size(current_gdbarch, SP_REGNUM));
	}
	return (cache);
}

static void
kgdb_trgt_trapframe_this_id(struct frame_info *next_frame, void **this_cache,
    struct frame_id *this_id)
{
	struct kgdb_frame_cache *cache;

	cache = kgdb_trgt_frame_cache(next_frame, this_cache);
	*this_id = frame_id_build(cache->sp, cache->pc);
}

static void
kgdb_trgt_trapframe_prev_register(struct frame_info *next_frame,
    void **this_cache, int regnum, int *optimizedp, enum lval_type *lvalp,
    CORE_ADDR *addrp, int *realnump, void *valuep)
{
	char dummy_valuep[MAX_REGISTER_SIZE];
	struct kgdb_frame_cache *cache;
	int ofs, regsz;
	static int ofs_fix = 0;
	static int ofs_fixed = 0;

	regsz = register_size(current_gdbarch, regnum);

	if (valuep == NULL)
		valuep = dummy_valuep;
	memset(valuep, 0, regsz);
	*optimizedp = 0;
	*addrp = 0;
	*lvalp = not_lval;
	*realnump = -1;

	if (!ofs_fixed) {
		/*
		 * In revision 1.117 of i386/i386/exception.S trap handlers
		 * were changed to pass trapframes by reference rather than
		 * by value.  Detect this by seeing if the first instruction
		 * at the 'calltrap' label is a "push %esp" which has the
		 * opcode 0x54.
		 */
		if (kgdb_parse("((char *)calltrap)[0]") == 0x54)
			ofs_fix = 4;
		ofs_fixed = 1;
	}
	ofs = (regnum >= I386_EAX_REGNUM && regnum <= I386_FS_REGNUM)
	    ? kgdb_trgt_frame_offset[regnum] + ofs_fix : -1;
	if (ofs == -1)
		return;

	cache = kgdb_trgt_frame_cache(next_frame, this_cache);
	switch (cache->frame_type) {
	case FT_NORMAL:
		break;
	case FT_INTRFRAME:
		ofs += 4;
		break;
	case FT_TIMERFRAME:
		break;
	case FT_INTRTRAPFRAME:
		ofs -= ofs_fix;
		break;
	default:
		fprintf_unfiltered(gdb_stderr, "Correct FT_XXX frame offsets "
		   "for %d\n", cache->frame_type);
		break;
	}
	*addrp = cache->sp + ofs;
	*lvalp = lval_memory;
	target_read_memory(*addrp, valuep, regsz);

	if (regnum == I386_EIP_REGNUM && *((int* )valuep) == 0)
	  {
	    /* When a trap is caused by a NULL pointer call, the EIP will be
	       set to 0. The return pointer is saved on the stack right 
	       before the trap frame. Because the trap hapened in the kernel,
	       the bottom portion of the trap frame structure (tf_esp and
	       tf_ss) will not be saved in the stack. See the "Trap handling"
	       comment in sys/i386/i386/exceptions.s for more details. The
	       last saved entry is tf_eflags and the return pointer follows
	       it in the location where ESP would have been. */
	    ofs = kgdb_trgt_frame_offset[I386_ESP_REGNUM] + ofs_fix;
	    *addrp = cache->sp + ofs;
	    target_read_memory(*addrp, valuep, regsz);
	  }
}

/* FIXME: is NORMAL_FRAME acceptable for this? */
/* FIXME: are the last three initialized correctly? */
static const struct frame_unwind kgdb_trgt_trapframe_unwind = {
    NORMAL_FRAME,
    &kgdb_trgt_trapframe_this_id,
    &kgdb_trgt_trapframe_prev_register,
    NULL, /*  const struct frame_data *unwind_data; */
    &kgdb_trgt_trapframe_sniffer, /*  frame_sniffer_ftype *sniffer;         */
    NULL  /*  frame_prev_pc_ftype *prev_pc;         */
};

const struct frame_unwind *
kgdb_trgt_trapframe_sniffer(struct frame_info *next_frame)
{
	char *pname;
	CORE_ADDR pc;

	pc = frame_pc_unwind(next_frame);
	pname = NULL;
	find_pc_partial_function(pc, &pname, NULL, NULL);
	if (pname == NULL)
		return (NULL);
	if (strcmp(pname, "dblfault_handler") == 0)
		return (&kgdb_trgt_dblfault_unwind);
	if (strcmp(pname, "calltrap") == 0 ||
	    (pname[0] == 'X' && pname[1] != '_'))
		return (&kgdb_trgt_trapframe_unwind);
	/* printf("%s: %llx =%s\n", __func__, pc, pname); */
	return (NULL);
}
