/*
 * Copyright (c) 2004 Marcel Moolenaar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * from: src/gnu/usr.bin/gdb/kgdb/trgt_alpha.c,v 1.2.2.1 2005/09/15 05:32:10 marcel
 */

#include <sys/cdefs.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/endian.h>
#include <machine/asm.h>
#include <err.h>
#include <kvm.h>
#include <string.h>

#include <defs.h>
#include <target.h>
#include <gdbthread.h>
#include <inferior.h>
#include <regcache.h>
#include <frame-unwind.h>
#include <mips-tdep.h>

#include "kgdb.h"

static CORE_ADDR mips_cpu_switch_addr = 0;
static size_t trapframe_size, trapframe_off, pcb_size, reg_size;
extern CORE_ADDR dumppcb;

#define REGCACHE_SUPPLY(regcache, regnum, reg_p) \
{ \
  isasize = register_size (current_gdbarch, regnum); \
  val = extract_signed_integer (reg_p, abisize); \
  store_signed_integer (buf, isasize, val); \
  regcache_raw_supply (regcache, regnum, buf); \
}

/* This is to account for STAND_ARG_SIZE in SAVE_REG (see exception.S) */
#define MIPS_TRAPFRAME_OFFSET 16

enum
{
  MIPS_AS_REGNUM = 1,
  MIPS_V1_REGNUM = 3,
  MIPS_A1_REGNUM = 5,
  MIPS_A2_REGNUM = 6,
  MIPS_A3_REGNUM = 7,
  MIPS_T0_REGNUM = 8,
  MIPS_T1_REGNUM = 9,
  MIPS_T2_REGNUM = 10,
  MIPS_T3_REGNUM = 11,
  MIPS_T4_REGNUM = 12,
  MIPS_T5_REGNUM = 13,
  MIPS_T6_REGNUM = 14,
  MIPS_T7_REGNUM = 15,
  MIPS_S0_REGNUM = 16,
  MIPS_S1_REGNUM = 17,
  MIPS_S2_REGNUM = 18,
  MIPS_S3_REGNUM = 19,
  MIPS_S4_REGNUM = 20,
  MIPS_S5_REGNUM = 21,
  MIPS_S6_REGNUM = 22,
  MIPS_S7_REGNUM = 23,
  MIPS_T8_REGNUM = 24,
  MIPS_K0_REGNUM = 26,
  MIPS_K1_REGNUM = 27,
  MIPS_GP_REGNUM = 28,
  MIPS_FP_REGNUM = 30,
  MIPS_SR_REGNUM = 32,
};

struct kgdb_frame_cache {
    CORE_ADDR	pc;
    CORE_ADDR	sp;
};

static int kgdb_trgt_frame_offset[40];

void
kgdb_trgt_fetch_registers(int regno __unused)
{
    struct kthr *kt;
    char* pcb;
    char* pcb_context;
    char* pc;
    char buf[MAX_REGISTER_SIZE];
    LONGEST val;
    size_t abisize, isasize, bias;

    kt = kgdb_thr_lookup_tid(ptid_get_pid(inferior_ptid));
    if (kt == NULL)
	return;

    abisize = mips_abi_regsize (get_regcache_arch (current_regcache));

    /*
     * After the kernel change to save 64bit registers in o32 we can't rely
     * on the compiled type of struct pcb. pcb_size and trapframe_size are
     * set dynamically to the actual sizes in the kernel. Trapframe is the
     * very first entry in pcb and pcb_context follows it. Register size is
     * set based on the size of the trapframe.
     */
    pcb = alloca(pcb_size);
    pcb_context = pcb + trapframe_size;
    if (kvm_read(kvm, kt->pcb, pcb, pcb_size) != pcb_size) {
	warnx("kvm_read: %s", kvm_geterr(kvm));
	memset(pcb, 0, pcb_size);
    }

    /*
     * Since we are only interested in the lower 32bits of every register,
     * all reads from pcb need to be offset by 4 when 64bit regs are used.
     */
    bias = (reg_size == 8 && abisize == 4) ? 4 : 0;
    pcb += bias;
    pcb_context += bias;

    REGCACHE_SUPPLY(current_regcache, MIPS_ZERO_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_ZERO_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_AS_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_AS_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_V0_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_V0_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_V1_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_V1_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_A0_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_A0_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_A1_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_A1_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_A2_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_A2_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_A3_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_A3_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T0_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T0_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T1_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T1_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T2_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T2_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T3_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T3_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T4_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T4_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T5_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T5_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T6_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T6_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T7_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T7_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T8_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T8_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_T9_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_T9_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_K0_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_K0_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_K1_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_K1_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_EMBED_LO_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_EMBED_LO_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_EMBED_HI_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_EMBED_HI_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_EMBED_BADVADDR_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_EMBED_BADVADDR_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_EMBED_CAUSE_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_EMBED_CAUSE_REGNUM]);
    REGCACHE_SUPPLY(current_regcache, MIPS_GP_REGNUM, 
		    pcb + kgdb_trgt_frame_offset[MIPS_GP_REGNUM]);

    if (kt->cpu == NOCPU || kt->pcb == dumppcb) {
	REGCACHE_SUPPLY(current_regcache, MIPS_S0_REGNUM, 
			pcb_context + reg_size * 0);
	REGCACHE_SUPPLY(current_regcache, MIPS_S1_REGNUM, 
			pcb_context + reg_size * 1);
	REGCACHE_SUPPLY(current_regcache, MIPS_S2_REGNUM, 
			pcb_context + reg_size * 2);
	REGCACHE_SUPPLY(current_regcache, MIPS_S3_REGNUM, 
			pcb_context + reg_size * 3);
	REGCACHE_SUPPLY(current_regcache, MIPS_S4_REGNUM, 
			pcb_context + reg_size * 4);
	REGCACHE_SUPPLY(current_regcache, MIPS_S5_REGNUM, 
			pcb_context + reg_size * 5);
	REGCACHE_SUPPLY(current_regcache, MIPS_S6_REGNUM, 
			pcb_context + reg_size * 6);
	REGCACHE_SUPPLY(current_regcache, MIPS_S7_REGNUM, 
			pcb_context + reg_size * 7);
	REGCACHE_SUPPLY(current_regcache, MIPS_SP_REGNUM, 
			pcb_context + reg_size * 8);
	REGCACHE_SUPPLY(current_regcache, MIPS_FP_REGNUM, 
			pcb_context + reg_size * 9);
	REGCACHE_SUPPLY(current_regcache, MIPS_RA_REGNUM, 
			pcb_context + reg_size * 10);
	REGCACHE_SUPPLY(current_regcache, MIPS_SR_REGNUM, 
			pcb_context + reg_size * 11);

	if (kt->pcb != dumppcb) {
	    pc = (abisize == 4 ? 4 : 0) + (char*) &mips_cpu_switch_addr;
	} else {
	    pc = pcb + kgdb_trgt_frame_offset[MIPS_EMBED_PC_REGNUM];
        }
        REGCACHE_SUPPLY(current_regcache, MIPS_EMBED_PC_REGNUM, pc);
    }
    else {
	REGCACHE_SUPPLY(current_regcache, MIPS_S0_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_S0_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_S1_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_S1_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_S2_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_S2_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_S3_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_S3_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_S4_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_S4_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_S5_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_S5_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_S6_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_S6_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_S7_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_S7_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_SP_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_SP_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_FP_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_FP_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_RA_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_RA_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_SR_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_SR_REGNUM]);
	REGCACHE_SUPPLY(current_regcache, MIPS_EMBED_PC_REGNUM, 
			pcb + kgdb_trgt_frame_offset[MIPS_EMBED_PC_REGNUM]);
    }
}

void
kgdb_trgt_store_registers(int regno __unused)
{

    fprintf_unfiltered(gdb_stderr, "XXX: %s\n", __func__);
}

void
kgdb_trgt_init_type_info(void)
{
	mips_cpu_switch_addr = kgdb_parse("mips_cpu_switch");

	/* 
	 * Initialize sizes of trapframe and pcb. Set register offsets
	 * based on the current size of the trap frame.
	 */
	trapframe_size = kgdb_parse("sizeof(struct trapframe)");
	pcb_size = kgdb_parse("sizeof(struct pcb)");
	reg_size = kgdb_parse("&((struct trapframe *)0)->ast");

	/*
	 * If trapframe offset is found in the kernel, use it
	 * Otherwise fall back to the hard coded value.
	 */
	errno = 0;
	trapframe_off = kgdb_parse("__mips_trapframe_offset");
	if (errno != 0)
	  trapframe_off = MIPS_TRAPFRAME_OFFSET;

	/*
	 * These are temp registers whose names can be used in either the old
	 * or new ABI, although they map to different physical registers.  In
	 * the old ABI, they map to t4-t7, and in the new ABI,they map to a4-a7.
	 *
	 * Because they overlap with the last 4 arg regs in the new ABI, ta0-ta3
	 * should be used only when we need more than t0-t3.
	 *
	 * Here's how a_/t_/ta_ register names map to physical registers in
	 * different ABIs:
	 *
	 *         |  4  5  6  7   8   9  10  11  12  13  14  15
	 * --------+--------------------------------------------
	 * o32/o64 | a0 a1 a2 a3  t0  t1  t2  t3  t4  t5  t6  t7
	 *         |                             ta0 ta1 ta2 ta3
	 * --------+--------------------------------------------
	 * n32/n64 | a0 a1 a2 a3  a4  a5  a6  a7  t0  t1  t2  t3
	 *         |             ta0 ta1 ta2 ta3 
	 * --------+---------------------------------------------
	 */

	kgdb_trgt_frame_offset[MIPS_ZERO_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->zero");
	kgdb_trgt_frame_offset[MIPS_AS_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->ast");
	kgdb_trgt_frame_offset[MIPS_V0_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->v0");
	kgdb_trgt_frame_offset[MIPS_V1_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->v1");
	kgdb_trgt_frame_offset[MIPS_A0_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->a0");
	kgdb_trgt_frame_offset[MIPS_A1_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->a1");
	kgdb_trgt_frame_offset[MIPS_A2_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->a2");
	kgdb_trgt_frame_offset[MIPS_A3_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->a3");

	if (kgdb_parse("&((struct trapframe *)0)->ta0") == 0) {
	    kgdb_trgt_frame_offset[MIPS_T0_REGNUM] = 
	        kgdb_parse("&((struct trapframe *)0)->t0");
	    kgdb_trgt_frame_offset[MIPS_T1_REGNUM] = 
	        kgdb_parse("&((struct trapframe *)0)->t1");
	    kgdb_trgt_frame_offset[MIPS_T2_REGNUM] = 
	        kgdb_parse("&((struct trapframe *)0)->t2");
	    kgdb_trgt_frame_offset[MIPS_T3_REGNUM] = 
	        kgdb_parse("&((struct trapframe *)0)->t3");
	    kgdb_trgt_frame_offset[MIPS_T4_REGNUM] = 
	        kgdb_parse("&((struct trapframe *)0)->t4");
	    kgdb_trgt_frame_offset[MIPS_T5_REGNUM] = 
	        kgdb_parse("&((struct trapframe *)0)->t5");
	    kgdb_trgt_frame_offset[MIPS_T6_REGNUM] = 
	        kgdb_parse("&((struct trapframe *)0)->t6");
	    kgdb_trgt_frame_offset[MIPS_T7_REGNUM] = 
	        kgdb_parse("&((struct trapframe *)0)->t7");
	}
	else {
	    if (mips_abi (current_gdbarch) == MIPS_ABI_N32 ||
		mips_abi (current_gdbarch) == MIPS_ABI_N64) {
	        kgdb_trgt_frame_offset[MIPS_T4_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->t0");
		kgdb_trgt_frame_offset[MIPS_T5_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->t1");
		kgdb_trgt_frame_offset[MIPS_T6_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->t2");
		kgdb_trgt_frame_offset[MIPS_T7_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->t3");
		kgdb_trgt_frame_offset[MIPS_T0_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->ta0");
		kgdb_trgt_frame_offset[MIPS_T1_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->ta1");
		kgdb_trgt_frame_offset[MIPS_T2_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->ta2");
		kgdb_trgt_frame_offset[MIPS_T3_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->ta3");
	    }
	    else {
		kgdb_trgt_frame_offset[MIPS_T0_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->t0");
		kgdb_trgt_frame_offset[MIPS_T1_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->t1");
		kgdb_trgt_frame_offset[MIPS_T2_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->t2");
		kgdb_trgt_frame_offset[MIPS_T3_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->t3");
		kgdb_trgt_frame_offset[MIPS_T4_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->ta0");
		kgdb_trgt_frame_offset[MIPS_T5_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->ta1");
		kgdb_trgt_frame_offset[MIPS_T6_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->ta2");
		kgdb_trgt_frame_offset[MIPS_T7_REGNUM] = 
		    kgdb_parse("&((struct trapframe *)0)->ta3");
	    }
	}

	kgdb_trgt_frame_offset[MIPS_S0_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s0");
	kgdb_trgt_frame_offset[MIPS_S1_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s1");
	kgdb_trgt_frame_offset[MIPS_S2_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s2");
	kgdb_trgt_frame_offset[MIPS_S3_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s3");
	kgdb_trgt_frame_offset[MIPS_S4_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s4");
	kgdb_trgt_frame_offset[MIPS_S5_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s5");
	kgdb_trgt_frame_offset[MIPS_S6_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s6");
	kgdb_trgt_frame_offset[MIPS_S7_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s7");
	kgdb_trgt_frame_offset[MIPS_T8_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->t8");
	kgdb_trgt_frame_offset[MIPS_T9_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->t9");
	kgdb_trgt_frame_offset[MIPS_K0_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->k0");
	kgdb_trgt_frame_offset[MIPS_K1_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->k1");
	kgdb_trgt_frame_offset[MIPS_GP_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->gp");
	kgdb_trgt_frame_offset[MIPS_SP_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->sp");
	kgdb_trgt_frame_offset[MIPS_FP_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->s8");
	kgdb_trgt_frame_offset[MIPS_RA_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->ra");
	kgdb_trgt_frame_offset[MIPS_SR_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->sr");
	kgdb_trgt_frame_offset[MIPS_EMBED_LO_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->mullo");
	kgdb_trgt_frame_offset[MIPS_EMBED_HI_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->mulhi");
	kgdb_trgt_frame_offset[MIPS_EMBED_BADVADDR_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->badvaddr");
	kgdb_trgt_frame_offset[MIPS_EMBED_CAUSE_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->cause");
	kgdb_trgt_frame_offset[MIPS_EMBED_PC_REGNUM] = 
	    kgdb_parse("&((struct trapframe *)0)->pc");
}

static struct kgdb_frame_cache *
kgdb_trgt_frame_cache(struct frame_info *next_frame, void **this_cache)
{
    char buf[MAX_REGISTER_SIZE];
    struct kgdb_frame_cache *cache;
    size_t abi_regsz, bias;
    struct frame_id nid;
    char* trapframe = alloca(trapframe_size);

    abi_regsz = mips_abi_regsize (current_gdbarch);

    /*
     * Since we are only interested in the lower 32bits of every register,
     * all reads need to be offset by 4 when 64bit regs are used.
     */
    bias = (reg_size == 8 && abi_regsz == 4) ? 4 : 0;
    cache = *this_cache;
    if (cache == NULL) {
	cache = FRAME_OBSTACK_ZALLOC(struct kgdb_frame_cache);
	*this_cache = cache;
	nid = get_frame_id(next_frame);
	cache->sp = nid.stack_addr; 

	/* read the pc value from trapframe */
	target_read_memory(cache->sp, trapframe, trapframe_size);
	cache->pc = 
	  extract_signed_integer(trapframe + bias + trapframe_off + 
				 MIPS_EMBED_PC_REGNUM * reg_size, abi_regsz); 
    }
    return (cache);
}

static void
kgdb_trgt_trapframe_this_id(struct frame_info *next_frame, void **this_cache,
			    struct frame_id *this_id)
{
    struct kgdb_frame_cache *cache;

    cache = kgdb_trgt_frame_cache(next_frame, this_cache);
    *this_id = frame_id_build(cache->sp, cache->pc);
}

static void
kgdb_trgt_trapframe_prev_register(struct frame_info *next_frame, void **this_cache,
				  int regnum, int *optimizedp, 
				  enum lval_type *lvalp, CORE_ADDR *addrp,
				  int *realnump, void *valuep)
{
    char dummy_valuep[MAX_REGISTER_SIZE];
    struct kgdb_frame_cache *cache;
    size_t ofs, abi_regsz, bias;

    abi_regsz = mips_abi_regsize (current_gdbarch);

    /*
     * Since we are only interested in the lower 32bits of every register,
     * all reads need to be offset by 4 when 64bit regs are used.
     */
    bias = (reg_size == 8 && abi_regsz == 4) ? 4 : 0;

    if (valuep == NULL)
	valuep = dummy_valuep;
    memset(valuep, 0, reg_size);
    *optimizedp = 0;
    *addrp = 0;
    *lvalp = not_lval;
    *realnump = -1;

    if (regnum >= NUM_REGS)
	regnum -= NUM_REGS;

    if (regnum == MIPS_EMBED_PC_REGNUM)
	ofs = trapframe_off + MIPS_EMBED_PC_REGNUM * reg_size;
    else 
	ofs = (regnum >= 0 && regnum <= MIPS_RA_REGNUM) ?
	  (trapframe_off + kgdb_trgt_frame_offset[regnum]) : -1;

    if (ofs == -1)
	return;

    cache = kgdb_trgt_frame_cache(next_frame, this_cache);

    /* addrp points to the register saved spot in trapframe */
    *addrp = cache->sp + ofs + bias;
    *lvalp = lval_memory;
    target_read_memory(*addrp, valuep, abi_regsz);

    if (regnum == MIPS_EMBED_PC_REGNUM) {
	/* Work around a frame unwind check */
	*lvalp = lval_register;
	*addrp = 0;
	*realnump = MIPS_EMBED_PC_REGNUM;
    }
}

/* FIXME: is NORMAL_FRAME acceptable for this? */
/* FIXME: are the last three initialized correctly? */
static const struct frame_unwind kgdb_trgt_trapframe_unwind = {
    NORMAL_FRAME,
    &kgdb_trgt_trapframe_this_id,
    &kgdb_trgt_trapframe_prev_register,
    0, /*  const struct frame_data *unwind_data; */
    0, /*  frame_sniffer_ftype *sniffer;         */
    0  /*  frame_prev_pc_ftype *prev_pc;         */
};

const struct frame_unwind *
kgdb_trgt_trapframe_sniffer(struct frame_info *next_frame)
{

    char *pname;
    CORE_ADDR pc;

    pc = frame_pc_unwind(next_frame);
    pname = NULL;
    find_pc_partial_function(pc, &pname, NULL, NULL);
    if (pname == NULL)
	return (NULL);
    if ((strcmp(pname, "MipsKernIntr") == 0) ||
	(strcmp(pname, "MipsKernGenException") == 0) ||
	(strcmp(pname, "MipsUserIntr") == 0) ||
	(strcmp(pname, "MipsUserGenException") == 0))
	return (&kgdb_trgt_trapframe_unwind);
    /* printf("%s: %llx =%s\n", __func__, pc, pname); */
    return (NULL);
}
